/*
  DiyouPCBFirm.h - DiyouPCB Config file

  Part of DiyouPCBFirm
  Based on Marlin & Grbl

Copyright (c) 2013 V.Ayllon.  All rights reserved.

DiyouPCBFirm is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
to anyone for the consequences of using it or for whether it serves any
particular purpose or works at all, unless he says so in writing.  Refer
to the GNU General Public License, Version 2 or later, for full details.

Everyone is granted permission to copy, modify and redistribute
DiyouPCBFirm, but only under the conditions described in the GNU General Public
License.  A copy of this license is supposed to have been
given to you along with DiyouPCB so you can know your rights and
responsibilities.  It should be in a file named COPYING.  Among other
things, the copyright notice and this notice must be preserved on all
copies.

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
#ifndef DIYOU_PCB_H
#define DIYOU_PCB_H
#define  FORCE_INLINE __attribute__((always_inline)) inline
#ifndef __AVR_ATmega328P__
	#define __AVR_ATmega328P__
#endif

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <inttypes.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include "fastio.h"
#include "Configuration.h"
#include "pins.h"
#include "serial.h"


#ifdef __cplusplus
#undef PROGMEM
#define PROGMEM __attribute__(( section(".progmem.data") ))
#undef PSTR
#endif

#ifndef PSTR
#define PSTR(s) (__extension__({static const char __c[] PROGMEM = (s);&__c[0];}))
#endif 

enum AxisEnum {X_AXIS=0, Y_AXIS=1};

// Bit field and masking macros
#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif
#define bit(n) (1 << n) 
#define bit_true(x,mask) (x |= mask)
#define bit_false(x,mask) (x &= ~mask)
#define bit_toggle(x,mask) (x ^= mask)
#define bit_istrue(x,mask) ((x & mask) != 0)
#define bit_isfalse(x,mask) ((x & mask) == 0)
// Define system executor bit map. Used internally by runtime protocol as runtime command flags, 
// which notifies the main program to execute the specified runtime command asynchronously.
// NOTE: The system executor uses an unsigned 8-bit volatile variable (8 flag limit.) The default
// flags are always false, so the runtime protocol only needs to check for a non-zero value to 
// know when there is a runtime command to execute.
#define EXEC_STATUS_REPORT  bit(0) // bitmask 00000001
#define EXEC_CYCLE_START    bit(1) // bitmask 00000010
#define EXEC_CYCLE_STOP     bit(2) // bitmask 00000100
#define EXEC_FEED_HOLD      bit(3) // bitmask 00001000
#define EXEC_RESET          bit(4) // bitmask 00010000
// Define system state bit map. The state variable primarily tracks the individual functions
// of Grbl to manage each without overlapping. It is also used as a messaging flag for
// critical events.
#define STATE_IDLE       0 // Must be zero.
#define STATE_ONLINE     1 // Initial power up state.
#define STATE_HOLD       2 // Executing feed hold
#define STATE_HELD       3 // Executing feed hold

// Define global system variables
typedef struct {
	uint8_t state;			      // The system state of the printer
	uint8_t execute;			  // Order execution
	float max_feedrate[NUM_AXIS]; // set the max speeds
	float axis_steps_per_unit[NUM_AXIS];
	unsigned long axis_steps_per_sqr_second[NUM_AXIS];
	uint16_t current_feed_rate;	     // Current focus position.
	uint8_t current_focus_pos;	     // Current focus position.
	uint8_t current_laser_power;	 // The current laser power
	uint8_t current_laser_state;	 // Current laser state ON/OFF
	float current_position[NUM_AXIS];// Current_position in mm
	float target[NUM_AXIS];			 // Target position in mm
	float offset[NUM_AXIS];			 // Offset in circles
    int8_t current_dir[NUM_AXIS];   // Current direction of travel
	float backlash[NUM_AXIS];		 // backlash comp. in mm
} system_t;

extern system_t sys;

#define ON	0x1
#define OFF 0x0

#define HIGH 0x1
#define LOW	 0x0

#define min(a,b) ((a)<(b)?(a):(b))
#define max(a,b) ((a)>(b)?(a):(b))

// Convert from degrees to radians
FORCE_INLINE float radians(float theta){
	return theta*(M_PI/180.);
}

#define SERIAL_PROTOCOL(x) print(x);
#define SERIAL_PROTOCOL_F(x,y) print(x,y);
#define SERIAL_PROTOCOLPGM(x) serialprintPGM(PSTR(x));
#define SERIAL_PROTOCOLLN(x) {print(x);serial_write('\n');}
#define SERIAL_PROTOCOLLNPGM(x) {serialprintPGM(PSTR(x));serial_write('\n');}


const char errormagic[] PROGMEM ="Error:";
const char echomagic[] PROGMEM ="echo:";
#define SERIAL_ERROR_START serialprintPGM(errormagic);
#define SERIAL_ERROR(x) SERIAL_PROTOCOL(x)
#define SERIAL_ERRORPGM(x) SERIAL_PROTOCOLPGM(x)
#define SERIAL_ERRORLN(x) SERIAL_PROTOCOLLN(x)
#define SERIAL_ERRORLNPGM(x) SERIAL_PROTOCOLLNPGM(x)

#define SERIAL_ECHO_START serialprintPGM(echomagic);
#define SERIAL_ECHO(x) SERIAL_PROTOCOL(x)
#define SERIAL_ECHOPGM(x) SERIAL_PROTOCOLPGM(x)
#define SERIAL_ECHOLN(x) SERIAL_PROTOCOLLN(x)
#define SERIAL_ECHOLNPGM(x) SERIAL_PROTOCOLLNPGM(x)


#define SERIAL_ECHOPAIR(name,value) (serial_echopair_P(PSTR(name),(value)))

void serial_echopair_P(const char *s_P, float v);
void serial_echopair_P(const char *s_P, double v);
void serial_echopair_P(const char *s_P, unsigned long v);

// Print a string stored in PGM-memory
FORCE_INLINE void serialprintPGM(const char *s)
{
  char c;
  while ((c = pgm_read_byte_near(s++)))
    serial_write(c);
}

#if MOTORS_ENABLE_PIN > -1
  #define enable_motors() WRITE(MOTORS_ENABLE_PIN, MOTORS_ENABLE_ON)
  #define disable_motors() WRITE(MOTORS_ENABLE_PIN,!MOTORS_ENABLE_ON)
#else
  #define enable_motors() ;
  #define disable_motors() ;
#endif

void execute_runtime_cmd();
void clearToSend();
void sendError();
void manageInactivity();
void checkAxesActivity();
void clamToSoftwareEnstops(float target[NUM_AXIS]);
void getCommand();
void homeAllAxis(bool focus);
void config_reset_default();
void process_hpgl_commands(char *line);
void prepare_line_move();
void prepare_arc_move();
void update_position();
void report_status();
void reset_printer();
void delay_us(uint32_t us);
void delay_ms(uint16_t ms);

#ifndef CRITICAL_SECTION_START
  #define CRITICAL_SECTION_START  unsigned char _sreg = SREG; cli();
  #define CRITICAL_SECTION_END    SREG = _sreg;
#endif //CRITICAL_SECTION_START

extern float min_pos[NUM_AXIS];
extern float max_pos[NUM_AXIS];

#endif
